from copy import deepcopy

from piperabm.infrastructure.samples.infrastructure_2 import (
    model as model_infrastructure,
)


model = deepcopy(model_infrastructure)
model.set_seed(2)
model.society.neighbor_radius = 270
model.society.generate(
    num=10,
    gini_index=0.45,
    average_resources={
        "food": 10,
        "water": 10,
        "energy": 10,
    },
    average_balance=100,
)
model.set_seed(None)


if __name__ == "__main__":
    print(model.society)
    print("society serialized: ", model.society.serialize())
    # model.show()



import numpy as np

from piperabm.tools.gini import gini
from piperabm.society.info import *


class Generate:
    """
    Generate agents
    """

    def generate(
        self,
        num: int = 1,
        gini_index: float = 0,
        average_resources: dict = {
            "food": 10,
            "water": 10,
            "energy": 10,
        },
        average_balance: float = 0,
    ):
        """
        Generate agents
        """
        distribution = gini.lognorm(gini_index)
        socioeconomic_status_values = distribution.rvs(sample_size=num, percision=0.03)
        homes_id = self.infrastructure.homes
        for value in socioeconomic_status_values:
            socioeconomic_status = float(value)
            home_id = int(np.random.choice(homes_id))
            resources = {}
            for name in average_resources:
                resources[name] = average_resources[name] * socioeconomic_status
            balance = average_balance * socioeconomic_status
            self.add_agent(
                home_id=home_id,
                socioeconomic_status=socioeconomic_status,
                resources=resources,
                balance=balance,
            )


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.set_seed(2)
    model.society.generate(gini_index=0.45, num=2, average_balance=1000)
    model.set_seed(None)

    print("gini index: ", model.society.gini_index)
    print("society serialized:\n", model.society.serialize())



from piperabm.tools.coordinate.projection.mercator import Mercator
from piperabm.tools.coordinate.projection.latlong_xyz import latlong_xyz, xyz_latlong
from piperabm.tools.coordinate import rotate


"""
Coordinate projection utilities converting between geographic (latitude, longitude) and Cartesian (x, y) coordinates.

This module uses 3D rotation and Mercator projection to:
- Convert latitude/longitude to local x/y (latlong_xy).
- Convert local x/y back to latitude/longitude (xy_latlong).

"""

EARTH_RADIUS = 6378000  # meters


def latlong_xy(
    latitude_0: float = 0.0,
    longitude_0: float = 0.0,
    latitude: float = 0.0,
    longitude: float = 0.0,
) -> tuple[float, float]:
    """
    Convert geographic coordinates to Cartesian (x, y) using a Mercator projection around a specified origin.

    :param float latitude_0: Latitude of the projection origin, in degrees.
    :param float longitude_0: Longitude of the projection origin, in degrees.
    :param float latitude: Latitude of the point to project, in degrees.
    :param float longitude: Longitude of the point to project, in degrees.
    :return: A tuple `(x, y)` of projected coordinates in meters.
    :rtype: tuple(float, float)
    """
    vector = latlong_xyz(latitude, longitude)
    vector = rotate.z(vector, longitude_0)
    vector = rotate.y(vector, -latitude_0)
    new_latitude, new_longitude = xyz_latlong(vector)
    x, y = Mercator.project(new_latitude, new_longitude, radius=EARTH_RADIUS)
    return x, y


def xy_latlong(
    latitude_0: float = 0.0, longitude_0: float = 0.0, x: float = 0.0, y: float = 0.0
) -> tuple[float, float]:
    """
    Convert Cartesian (x, y) back to geographic coordinates around a specified origin using the inverse Mercator projection.

    :param float latitude_0: Latitude of the projection origin, in degrees.
    :param float longitude_0: Longitude of the projection origin, in degrees.
    :param float x: X coordinate in meters relative to the projection origin.
    :param float y: Y coordinate in meters relative to the projection origin.
    :return: A tuple `(latitude, longitude)` of geographic coordinates in degrees.
    :rtype: tuple(float, float)
    """
    new_latitude, new_longitude = Mercator.inverse(x, y, radius=EARTH_RADIUS)
    vector = latlong_xyz(new_latitude, new_longitude)
    vector = rotate.y(vector, latitude_0)
    vector = rotate.z(vector, -longitude_0)
    latitude, longitude = xyz_latlong(vector)
    return latitude, longitude


def _example():  # pragma: no cover
    latitude_0 = 70
    longitude_0 = -150

    latitude = latitude_0 + 1
    longitude = longitude_0 + 1

    x, y = latlong_xy(latitude_0, longitude_0, latitude, longitude)
    print(f"x, y: {x}, {y}")

    latitude, longitude = xy_latlong(latitude_0, longitude_0, x, y)
    print(f"latitude, longitude: {latitude}, {longitude}")


if __name__ == "__main__":
    _example()



import os
import matplotlib.pyplot as plt
from copy import deepcopy

from piperabm.tools.average import average as avg


class Accessibility:
    """
    Manage accessibility measurement
    """

    type = "accessibility"

    def __init__(self, measurement):
        self.measurement = measurement
        self.values = {}
        self.resource_names = ["food", "water", "energy"]

    def add(self, id: int, value: float) -> None:
        """
        Add new accessibility value
        """
        if not id in self.values:
            self.values[id] = []
        self.values[id].append(value)

    def get(self, agent: int, resource: str, time_step: int):
        """
        Get the desired entry
        """
        return self.values[agent][time_step][resource]

    @property
    def agents(self):
        """
        Return a list of agents id
        """
        return list(self.values.keys())

    def rearrange(self, agents="all", resources="all", _from=None, _to=None):
        """
        Return accessibility values for all resources for across agents.

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        if agents == "all":
            ids = self.agents
        elif isinstance(agents, list):
            ids = agents
        elif isinstance(agents, int):
            ids = [agents]
        elif agents is None:
            ids = self.agents
        else:
            raise ValueError

        if resources == "all":
            resource_names = self.resource_names
        elif isinstance(resources, list):
            resource_names = resources
        elif isinstance(resources, str) and resources != "all":
            resource_names = [resources]
        elif resources is None:
            resource_names = self.resource_names
        else:
            raise ValueError

        if _from is None:
            _from = 0

        if _to is None:
            _to = self.len

        result = {}
        for agent_id in ids:
            result[agent_id] = {}
            for resource_name in resource_names:
                values = []
                for i in range(_from, _to):
                    value = self.get(
                        agent=agent_id, resource=resource_name, time_step=i
                    )
                    values.append(value)
                result[agent_id][resource_name] = values

        return result

    @property
    def len(self) -> int:
        """
        Return total number of entries
        """
        return self.measurement.len

    def sum_resources(
        self, agents="all", resources="all", _from=None, _to=None
    ) -> dict:
        r"""
        Sum the accessibility to resources.

        For each agent *i*, time *t*, and resource *r*:

        .. math::

            A_{i,t,r} = \frac{R_{i,t}}{R^{\max}_i}

        To aggregate across resources:

        .. math::

            A_{i,t} = \left(\prod_{r=1}^R A_{i,t,r}\right)^{1/R}

        The reason is, if any of the resources reach zero, life will become challenging.

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        data = self.rearrange(agents=agents, resources=resources, _from=_from, _to=_to)
        if _from is None:
            _from = 0
        if _to is None:
            _to = self.len
        length = _to - _from
        result = {}
        for agent_id in data:
            result[agent_id] = []
            agent_data = data[agent_id]
            for i in range(length):
                vals = []
                for resource_name in agent_data:
                    acc_vals = agent_data[resource_name]
                    vals.append(acc_vals[i])
                result[agent_id].append(avg.geometric(values=vals))
        return result

    def sum_agents(self, agents="all", resources="all", _from=None, _to=None) -> list:
        r"""
        Sum of accessibility for agents.

        For each agent *i*, time *t*, and resource *r*:

        .. math::

            A_{i,t,r} = \frac{R_{i,t}}{R^{\max}_i}

        Community average at time *t*:

        .. math::
            A_{t,r} = \frac{1}{N}\sum_{i=1}^N A_{i,t,r}

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        data = self.sum_resources(
            agents=agents, resources=resources, _from=_from, _to=_to
        )
        if _from is None:
            _from = 0
        if _to is None:
            _to = self.len
        length = _to - _from
        result = []
        for i in range(length):
            vals = []
            for agent_id in data:
                agent_data = data[agent_id]
                vals.append(agent_data[i])
            result.append(avg.arithmetic(values=vals))
        return result

    def __call__(self, agents="all", resources="all", _from=None, _to=None) -> list:
        return self.sum_agents(agents=agents, resources=resources, _from=_from, _to=_to)

    def average(self, agents="all", resources="all", _from=None, _to=None) -> float:
        r"""
        Calculate total average.

        For each agent *i*, time *t*, and resource *r*:

        .. math::

            A_{i,t,r} = \frac{R_{i,t}}{R^{\max}_i}

        To aggregate across resources:

        .. math::

            A_{i,t} = \left(\prod_{r=1}^R A_{i,t,r}\right)^{1/R}

        Community average at time *t*:

        .. math::
            A_t = \frac{1}{N}\sum_{i=1}^N A_{i,t}

        And over the full duration *T*:

        .. math::
            A = \frac{\int_0^T A_t\,dt}{\int_0^T A_\max\,dt}

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        values = self.__call__(agents=agents, resources=resources, _from=_from, _to=_to)
        weights = self.measurement.delta_times(_from=_from, _to=_to)
        result = avg.arithmetic(values=values, weights=weights)
        if isinstance(result, complex):
            result = float(result.real)
        return result

    def create_plot(
        self, agents="all", resources="all", _from=None, _to=None, info=None
    ):
        """
        Create plot for accessibility.

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        fig, ax = plt.subplots()
        title = "Accessibility"
        ylabel = deepcopy(title)
        if info is not None:
            title += info
        ax.set_title(title)
        xs = self.measurement.filter_times(_from=_from, _to=_to)
        yx = self.__call__(agents=agents, resources=resources, _from=_from, _to=_to)
        ax.plot(xs, yx, color="blue")
        ax.set_xlabel("Time")
        ax.set_ylabel(ylabel)
        return fig

    def show(self, agents="all", resources="all", _from=None, _to=None, info=None):
        """
        Draw plot for accessibility over time.

        Parameters
        ----------
        agents : default='all'
            Accepts either 'all', single id of an agent, or a list of agents id. These are the agents that their accessibility values will be calulcated as sum.
        resources : default='all'
            Accepts either 'all', name of a single resource, or a list of resource names. These are the resources that accessibility to them will be calculated.
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        fig = self.create_plot(
            agents=agents, resources=resources, _from=_from, _to=_to, info=info
        )
        plt.show()

    def save(self, agents="all", resources="all", _from=None, _to=None, info=None):
        """
        Save plot
        """
        fig = self.create_plot(
            agents=agents, resources=resources, _from=_from, _to=_to, info=info
        )
        path = self.measurement.result_directory
        filepath = os.path.join(path, self.type)
        fig.savefig(filepath)

    def serialize(self) -> dict:
        """
        Serialize
        """
        return {"values": self.values, "type": self.type}

    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        self.values = data["values"]


if __name__ == "__main__":

    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour)  # Base

    # 1
    measure.add_time(value=1 * hour)
    measure.accessibility.add(id=1, value={"food": 1, "water": 1, "energy": 1})
    measure.accessibility.add(id=2, value={"food": 0.8, "water": 0.7, "energy": 0.6})
    # 2
    measure.add_time(value=2 * hour)
    measure.accessibility.add(id=1, value={"food": 0.9, "water": 0.8, "energy": 0.7})
    measure.accessibility.add(id=2, value={"food": 0.5, "water": 0.6, "energy": 0.4})
    # 3
    measure.add_time(value=3 * hour)
    measure.accessibility.add(id=1, value={"food": 0.8, "water": 0.7, "energy": 0.6})
    measure.accessibility.add(id=2, value={"food": 0.2, "water": 0.4, "energy": 0.3})
    # 4
    measure.add_time(value=4 * hour)
    measure.accessibility.add(id=1, value={"food": 0.7, "water": 0.6, "energy": 0.5})
    measure.accessibility.add(id=2, value={"food": 0, "water": 0.3, "energy": 0.2})
    # 5
    measure.add_time(value=5 * hour)
    measure.accessibility.add(id=1, value={"food": 0.6, "water": 0.5, "energy": 0.4})
    measure.accessibility.add(id=2, value={"food": 0, "water": 0.3, "energy": 0.2})

    agents = "all"
    resources = "all"
    _from = None
    _to = None
    print(
        "accessibilities: ",
        measure.accessibility(agents=agents, resources=resources, _from=_from, _to=_to),
    )
    print(
        "average: ",
        measure.accessibility.average(
            agents=agents, resources=resources, _from=_from, _to=_to
        ),
    )
    measure.accessibility.show(agents=agents, resources=resources, _from=_from, _to=_to)



from .infrastructure import Infrastructure

# Bootstraping for DecisionMaking
# Source: https://github.com/aslan-ng/python-bootstrapper
import os, sys, inspect, importlib.util

# 1) import the “official” DecisionMaking
from .degradation import Degradation as _BaseDegradation


def _discover_local(plugin_name: str):
    # find the folder containing the user's entry-point script
    main_mod = sys.modules.get("__main__")
    if hasattr(main_mod, "__file__"):
        script_dir = os.path.dirname(os.path.abspath(main_mod.__file__))
    else:
        script_dir = os.getcwd()

    # look for a file named decision_making.py next to it
    plugin_path = os.path.join(script_dir, f"{plugin_name}.py")
    if not os.path.isfile(plugin_path):
        return _BaseDegradation

    # dynamically load it as a module
    spec = importlib.util.spec_from_file_location(plugin_name, plugin_path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    # find any class in that module that subclasses our base
    for obj in vars(mod).values():
        if (
            inspect.isclass(obj)
            and issubclass(obj, _BaseDegradation)
            and obj is not _BaseDegradation
        ):
            return obj

    # fallback if nothing suitable was found
    return _BaseDegradation


# 2) override the name in our package
Degradation = _discover_local("degradation")



from copy import deepcopy

from piperabm.economy.trade import MultiResourceTrade
from piperabm.economy.trade.allocation import find_exchanges


class Trade:
    """
    Manage trade between players when updating
    """

    def trade(self, agents: list = [], markets: list = []):
        """
        Trade
        """
        names = self.resource_names
        players = []

        # Load agent players into solver
        for agent in agents:
            resources = {}
            enough_resources = {}
            for name in names:
                resources[name] = self.society.get_resource(id=agent, name=name)
                enough_resources[name] = self.society.get_enough_resource(
                    id=agent, name=name
                )
            player = {
                "id": agent,
                "type": "agent",
                "resources": resources,
                "enough_resources": enough_resources,
                "balance": self.society.get_balance(id=agent),
            }
            players.append(player)

        # Load market players into solver
        for market in markets:
            resources = {}
            enough_resources = {}
            for name in names:
                resources[name] = self.infrastructure.get_resource(id=market, name=name)
                enough_resources[name] = self.infrastructure.get_enough_resource(
                    id=market, name=name
                )
            player = {
                "id": market,
                "type": "market",
                "resources": resources,
                "enough_resources": enough_resources,
                "balance": self.infrastructure.get_balance(id=market),
            }
            players.append(player)

        # Solve
        players_initial = deepcopy(players)
        players = MultiResourceTrade.solve(players=players, prices=self.prices)
        transactions = find_exchanges(
            players_initial=players_initial, players_final=players
        )

        # Update values
        for player in players:
            # Update agent players
            if player["type"] == "agent":
                for name in names:
                    self.society.set_resource(
                        id=player["id"], name=name, value=player["resources"][name]
                    )
                self.society.set_balance(id=player["id"], value=player["balance"])
            # Update market players
            elif player["type"] == "market":
                for name in names:
                    self.infrastructure.set_resource(
                        id=player["id"], name=name, value=player["resources"][name]
                    )
                self.infrastructure.set_balance(
                    id=player["id"], value=player["balance"]
                )
        # print(transactions)
        return flatten_transactions(transactions)


def flatten_transactions(transactions):
    results = []
    for resource_name in transactions:
        for transaction in transactions[resource_name]:
            result = [
                transaction[0],  # from
                transaction[1],  # to
                transaction[2],  # amount
                resource_name,  # resource name
            ]
            results.append(result)
    # print(results)
    return results


if __name__ == "__main__":

    from piperabm.society.samples.society_0 import model

    model.society.average_income = 0
    agents = model.society.agents

    wealth_0 = model.society.wealth(agents[0])
    wealth_1 = model.society.wealth(agents[1])
    if wealth_0 < wealth_1:
        id_low = agents[0]  # Agent with lower wealth
        id_high = agents[1]  # Agent with higher wealth
    food = model.society.get_resource(id_low, "food")
    model.society.set_resource(id_low, "food", value=food / 10)
    water = model.society.get_resource(id_low, "water")
    model.society.set_resource(id_low, "water", value=water / 5)

    print("\n" + ">>> Initial:")
    for agent in agents:
        balance = model.society.get_balance(agent)
        food = model.society.get_resource(agent, "food")
        water = model.society.get_resource(agent, "water")
        energy = model.society.get_resource(agent, "energy")
        print(
            f"id: {agent}, balance: {balance}, food: {food}, water: {water}, energy: {energy}"
        )

    transactions = model.update(duration=1)

    print("\n" + ">>> " + "Exchanges: ")
    for transaction in transactions:
        print(
            f"from: {transaction[0]}, to: {transaction[1]}, amount: {transaction[2]}, resource: {transaction[3]}"
        )

    print("\n" + ">>> Final:")
    for agent in agents:
        balance = model.society.get_balance(agent)
        food = model.society.get_resource(agent, "food")
        water = model.society.get_resource(agent, "water")
        energy = model.society.get_resource(agent, "energy")
        print(
            f"id: {agent}, balance: {balance}, food: {food}, water: {water}, energy: {energy}"
        )



"""
Simple world
"""

import piperabm as pa


model = pa.Model()
model.set_seed(2)
model.infrastructure.coeff_usage = 1
model.infrastructure.coeff_age = 1
model.infrastructure.add_market(
    pos=[40, -40],
    name="market",
    id=0,
    resources={"food": 100, "water": 100, "energy": 100},
)
model.infrastructure.add_home(pos=[-60, 40], name="home 1", id=1)
model.infrastructure.add_home(pos=[200, 20], name="home 2", id=2)
model.infrastructure.add_home(pos=[100, -180], name="home 3", id=3)
model.infrastructure.add_street(pos_1=[-60, 40], pos_2=[0, 0], name="street 1")
model.infrastructure.add_street(pos_1=[0, 0], pos_2=[80, 60], name="street 2")
model.infrastructure.add_street(pos_1=[80, 60], pos_2=[200, 20], name="street 3")
model.infrastructure.add_street(pos_1=[0, 0], pos_2=[100, -180], name="street 4")
model.bake()
model.set_seed(None)


if __name__ == "__main__":
    model.infrastructure.show()



import os

from piperabm.model.measurement import Measurement


path = os.path.dirname(os.path.realpath(__file__))

measure = Measurement(path=path, name="model")
hour = 3600
measure.add_time(0 * hour)  # Base

# 1
measure.add_time(value=1 * hour)
measure.add_accessibility(id=1, value={"food": 1, "water": 1, "energy": 1})
measure.add_accessibility(id=2, value={"food": 0.8, "water": 0.7, "energy": 0.6})
measure.add_travel_distance(value=1.1)
# 2
measure.add_time(value=2 * hour)
measure.add_accessibility(id=1, value={"food": 0.9, "water": 0.8, "energy": 0.7})
measure.add_accessibility(id=2, value={"food": 0.5, "water": 0.6, "energy": 0.4})
measure.add_travel_distance(value=0.9)
# 3
measure.add_time(value=3 * hour)
measure.add_accessibility(id=1, value={"food": 0.8, "water": 0.7, "energy": 0.6})
measure.add_accessibility(id=2, value={"food": 0.2, "water": 0.4, "energy": 0.3})
measure.add_travel_distance(value=0.3)
# 4
measure.add_time(value=4 * hour)
measure.add_accessibility(id=1, value={"food": 0.7, "water": 0.6, "energy": 0.5})
measure.add_accessibility(id=2, value={"food": 0, "water": 0.3, "energy": 0.2})
measure.add_travel_distance(value=0.46)
# 5
measure.add_time(value=5 * hour)
measure.add_accessibility(id=1, value={"food": 0.6, "water": 0.5, "energy": 0.4})
measure.add_accessibility(id=2, value={"food": 0, "water": 0.3, "energy": 0.2})
measure.add_travel_distance(value=0.2)

measure.save()



"""
Solves multi-resources for multiple players at fixed price
"""

from piperabm.economy.trade.nash_bargaining import NashBargaining as nb


class MultiResourceTrade:

    def resource_names(players: list):
        return players[0]["resources"].keys()

    def balance_allocations(players: list, prices: dict) -> list:
        # Balance allocation
        resource_names = MultiResourceTrade.resource_names(players=players)
        result = []
        for player in players:
            balance_allocation = {}
            needs = {}
            for resource_name in resource_names:
                need = max(
                    0,
                    (
                        player["enough_resources"][resource_name]
                        - player["resources"][resource_name]
                    )
                    * prices[resource_name],
                )
                needs[resource_name] = need
            all_needs = sum(needs.values())
            for resource_name in resource_names:
                if all_needs != 0:
                    balance_allocation[resource_name] = needs[resource_name] / all_needs
                else:
                    balance_allocation[resource_name] = 0
            result.append(balance_allocation)
        return result

    def prepare(players: list, balance_allocations: list, resource_name: str):
        """
        Extract the player info
        """
        result = []
        for i, player in enumerate(players):
            new = {}
            new["id"] = player["id"]
            new["resource"] = player["resources"][resource_name]
            new["enough_resource"] = player["enough_resources"][resource_name]
            new["balance"] = player["balance"] * balance_allocations[i][resource_name]
            result.append(new)
        return result

    """
    def transactions(players: list, prices: dict):
        balance_allocations = MultiResourceTrade.balance_allocations(players)
        resource_names = MultiResourceTrade.resource_names(players=players)
        all_transactions = {}
        for resource_name in resource_names:
            market_players = MultiResourceTrade.prepare(
                players=players,
                balance_allocations=balance_allocations,
                resource_name=resource_name
            )
            transactions = nb.transactions(players=market_players, price=prices[resource_name])
            all_transactions[resource_name] = transactions
            #players = nb.apply(players, transactions)
        return all_transactions
    """

    def sorted_markets(players, prices):
        """
        Sorted name of markets based on their size
        """
        market_sizes = {}
        for resource_name in MultiResourceTrade.resource_names(players):
            market_sizes[resource_name] = 0
        for player in players:
            resources = player["resources"]
            for resource_name in resources:
                market_sizes[resource_name] += resources[resource_name]
        for resource_name in market_sizes:
            market_sizes[resource_name] *= prices[resource_name]
        return sorted(market_sizes, key=market_sizes.get, reverse=True)

    def transactions(players: list, prices: dict) -> dict:
        """
        Calculate the proper transactions
        """
        markets = MultiResourceTrade.sorted_markets(players=players, prices=prices)
        all_transactions = {}
        for market in markets:
            balance_allocations = MultiResourceTrade.balance_allocations(
                players=players, prices=prices
            )
            market_players = MultiResourceTrade.prepare(
                players=players,
                balance_allocations=balance_allocations,
                resource_name=market,
            )
            transactions = nb.transactions(players=market_players, price=prices[market])
            all_transactions[market] = transactions
        return all_transactions

    def apply(players, transactions) -> list:
        """
        Apply transactions to players
        """
        num_players = len(players)
        resource_names = MultiResourceTrade.resource_names(players=players)
        for resource_name in resource_names:
            resource_transactions = transactions[resource_name]
            for i in range(num_players):
                players[i]["balance"] = (
                    players[i]["balance"] + resource_transactions["money"][i]
                )
                players[i]["resources"][resource_name] = (
                    players[i]["resources"][resource_name]
                    + resource_transactions["resource"][i]
                )
        return players

    def solve(players: list, prices: dict) -> list:
        """
        Solve until convergence
        """
        i = 0
        # max_itteration = 1
        max_itteration = len(players) ** 2
        while i < max_itteration:
            transactions = MultiResourceTrade.transactions(
                players=players, prices=prices
            )
            # print(transactions)
            if MultiResourceTrade.check_empty(transactions) is True:
                break
            else:
                players = MultiResourceTrade.apply(
                    players=players, transactions=transactions
                )
            i += 1
        """
        transactions = MultiResourceTrade.transactions(players=players, prices=prices)
        players = MultiResourceTrade.apply(players=players, transactions=transactions)
        """
        return players

    def check_empty(transactions: dict) -> bool:
        result = None
        threashold = 0
        transfers = 0
        for resource_name in transactions:
            transfers += sum(transactions[resource_name]["money"])
        if transfers <= threashold:
            result = True
        else:
            result = False
        return result


if __name__ == "__main__":
    prices = {"food": 10, "water": 10, "energy": 10}
    player_1 = {
        "id": 1,
        "type": "agent",
        "resources": {"food": 9, "water": 2, "energy": 3},
        "enough_resources": {"food": 10, "water": 10, "energy": 10},
        "balance": 100,
    }
    player_2 = {
        "id": 2,
        "type": "agent",
        "resources": {"food": 10, "water": 3, "energy": 10},
        "enough_resources": {"food": 10, "water": 10, "energy": 10},
        "balance": 100,
    }
    player_3 = {
        "id": 3,
        "type": "agent",
        "resources": {"food": 3, "water": 10, "energy": 7},
        "enough_resources": {"food": 10, "water": 10, "energy": 10},
        "balance": 10,
    }
    players = [player_1, player_2, player_3]

    print("Solves multi-resources for multiple players at fixed price.")

    # Initial
    print("\n" + ">>> " + "Initial: ")
    for player in players:
        print(player["resources"], ", balance:", player["balance"])

    # Solve
    players = MultiResourceTrade.solve(players=players, prices=prices)

    # Final
    print("\n" + ">>> " + "Final: ")
    for player in players:
        print(player["resources"], ", balance:", player["balance"])



import random


class Triangle:
    """
    Represent a single triangle for meshing purpose.

    Parameters
    ----------
    point_1 : list
        `x` and `y` coordinates of the vertice 1.
    point_2 : list
        `x` and `y` coordinates of the vertice 1.
    point_3 : list
        `x` and `y` coordinates of the vertice 1.
    density : float, default=1
        The higher the density is, the higher the chance of containing the point is.
    """

    def __init__(self, point_1: list, point_2: list, point_3: list, density: float = 1):
        self.A = point_1
        self.B = point_2
        self.C = point_3
        self.density = density

    @property
    def area(self):
        """
        Calculate the area of the triangle.
        """
        x1, y1 = self.A
        x2, y2 = self.B
        x3, y3 = self.C
        return abs(x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)) / 2.0

    def random_point(self) -> list:
        """
        Generate a random point inside the triangle using Barycentric coordinates.
        """
        r1, r2 = random.random(), random.random()
        sqrt_r1 = r1**0.5
        a = 1 - sqrt_r1
        b = sqrt_r1 * (1 - r2)
        c = r2 * sqrt_r1
        # Calculate the random point's coordinates
        x = a * self.A[0] + b * self.B[0] + c * self.C[0]
        y = a * self.A[1] + b * self.B[1] + c * self.C[1]
        return [x, y]

    @property
    def weight(self) -> float:
        """
        Both area and density are proportional to the probability of containing a point inside. Therefore, weight is defined as a more suitable parameter for point generation.
        """
        return self.area * self.density


if __name__ == "__main__":
    pos_1 = [0, 0]
    pos_2 = [2, 0]
    pos_3 = [0, 1]
    triangle = Triangle(pos_1, pos_2, pos_3)
    print(triangle.area)



import os
import csv
import keepdelta as kd

from piperabm.tools.json_file import JsonFile


class File:
    """
    Manages file handling methods of model
    """

    @property
    def result_directory(self):
        """
        Return result directory
        """
        if self.path is None:
            raise ValueError("Define path to continue.")
        result = os.path.join(self.path, "result")
        if self.name != "":
            result = os.path.join(result, self.name)
        return result

    def save(self, name: str) -> None:
        """
        Save model to file
        """
        path = self.result_directory
        if not os.path.exists(path):  # Create result directory if doesn't exist
            os.makedirs(path)
        data = self.serialize()
        file = JsonFile(path=path, filename=name)
        file.save(data)

    def save_initial(self) -> None:
        """
        Save initial state of model to file
        """
        self.save(name="initial")

    def save_final(self) -> None:
        """
        Save final state of model to file
        """
        self.save(name="final")

    def load(self, name: str) -> None:
        """
        Load model from file
        """
        file = JsonFile(path=self.result_directory, filename=name)
        if file.exists() is False:
            raise ValueError(f"File {name} doesn't exist in {self.result_directory}")
        data = file.load()
        self.deserialize(data)

    def load_initial(self) -> None:
        """
        Load initial state of model from file
        """
        self.load(name="initial")

    def load_final(self) -> None:
        """
        Load final state of model from file
        """
        self.load(name="final")

    def load_deltas(self, _from: int = None, _to: int = None) -> list:
        """
        Load all detlas from file
        """
        deltas_file = JsonFile(path=self.result_directory, filename="simulation")
        deltas = deltas_file.load()
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(deltas)
        deltas = deltas[_from:_to]
        return deltas

    def append_delta(self, delta) -> None:
        """
        Append the new delta to file
        """
        deltas_file = JsonFile(path=self.result_directory, filename="simulation")
        if deltas_file.exists() is False:
            deltas_file.save(data=[])
        deltas_file.append(delta)

    def apply_delta(self, delta) -> None:
        """
        Update model by applying a delta
        """
        self.deserialize(kd.apply(old=self.serialize(), delta=delta))

    def apply_deltas(self, deltas: list = None) -> None:
        """
        Update model by applying all deltas
        """
        if deltas is None:
            deltas = self.load_deltas()
        for delta in deltas:
            self.apply_delta(delta)

    def push(self, steps: int = 1) -> None:
        """
        Push model forward using deltas
        """
        current = self.step
        deltas = self.load_deltas(_from=current, _to=current + steps)
        self.apply_deltas(deltas=deltas)

    def append_transactions(self, transactions) -> None:
        """
        Append the new delta to file
        """
        name = "transactions" + "." + "csv"
        filepath = os.path.join(self.result_directory, name)
        headers = ["from", "to", "amount", "resource", "time"]
        if not os.path.exists(filepath):
            # Create it with headers only
            with open(filepath, mode="w", newline="") as file:
                writer = csv.writer(file)
                writer.writerow(headers)
        with open(filepath, mode="a", newline="") as file:
            writer = csv.writer(file)
            for transaction in transactions:
                writer.writerow(transaction)


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    model.path = os.path.dirname(os.path.realpath(__file__))
    model.save_initial()
    model.load_initial()



"""
Default values
"""

from copy import deepcopy


""" Idle consumption """
idle_resource_rates = {
    "food": 2 / (60 * 60 * 24),  # kg/day to kg/s
    "water": 2 / (60 * 60 * 24),  # kg/day to kg/s
    "energy": 2 / (60 * 60 * 24),  # kg/day to kg/s
}

""" Walk """
speed = 5 * ((1000) / (60 * 60))  # km/hour to m/s
transportation_resource_rates = {
    "food": 2 / (60 * 60 * 24),  # kg/day to kg/s
    "water": 1 / (60 * 60 * 24),  # kg/day to kg/s
    "energy": 1 / (60 * 60 * 24),  # kg/day to kg/s
}

""" Resource """
average_resources = {
    "food": 20,  # kg
    "water": 20,  # kg
    "energy": 20,  # kg
}
average_enough_resources = deepcopy(average_resources)



import numpy as np

from piperabm.model.serialize import Serialize
from piperabm.model.file import File
from piperabm.model.update import Update
from piperabm.model.graphics import Graphics
from piperabm.infrastructure import Infrastructure
from piperabm.society import Society
from piperabm.tools.symbols import SYMBOLS


class Model(Serialize, File, Update, Graphics):
    """
    Main class of simulation.

    Parameters
    ----------
    name : str, optional
        A label to distinguish this model and its results when running multiple instances. In single runs, this can left empty.
    prices : dict, optional
        Mapping of resource names to unit costs. FEWS Nexus framework expects `'food'`, `'water'`, `'energy'`.
    path : str, optional
        Directory for saving/loading simulation.
    seed : int, optional
        Integer seed for reproducibility.
    """

    type = "model"

    def __init__(
        self,
        name: str = "",
        prices: dict = {
            "food": 1,
            "water": 1,
            "energy": 1,
        },
        path=None,
        seed: int = None,
    ):
        """
        Initialize the Model.
        """
        super().__init__()
        self.time = 0
        self.step = 0
        self.infrastructure = Infrastructure()
        self.infrastructure.model = self  # Binding
        self.society = Society()
        self.society.model = self  # Binding
        self.name = name
        self.prices = prices
        self.path = path  # File saving
        self.set_seed(seed=seed)

    def set_seed(self, seed: int = None):
        """
        Set random generator seed for result reproducability.
        """
        self.seed = seed
        np.random.seed(seed)

    @property
    def resource_names(self):
        """
        Return name of resources in the model.
        """
        return list(self.prices.keys())

    def bake(
        self,
        save: bool = False,
        proximity_radius: float = SYMBOLS["eps"],
        search_radius: float = None,
        report: bool = False,
    ):
        """
        Prepare the model for the first simulation step.
        This will generate nodes/edges, and compute any necessary initial calculations to create a physically sensinble network.

        Parameters
        ----------
        save : bool, default=False
            If True, immediately serialize the infrastructure state to disk.
        proximity_radius : float, default=0
            The grammar rules use this value (in model units) to determine how close the elements should be to each other to impact each other, such as getting merged or split.
        search_radius : float or None, default=None
            Home and market nodes need to get connected to the street network. This is done using "Neighborhood access" edges. The grammar rule for this process is computationally expensive therefore setting a search radius can speed up the process specially in large and intricate networks. If set to `None`, all possible elements are network are evaluated.
        report : bool, default=False
            If `True`, reports the steps taken during the baking process, which can be useful for debugging or understanding the model.
        """
        self.infrastructure.bake(
            report=report,
            proximity_radius=proximity_radius,
            search_radius=search_radius,
        )
        if save is True:
            self.save(state="infrastructure")


if __name__ == "__main__":

    from piperabm.model import Model

    model = Model()
    model.infrastructure.add_home(pos=[0, 0])
    model.infrastructure.add_street(pos_1=[-5, 0], pos_2=[5, 0])
    model.bake()
    print(model.infrastructure)



class Degradation:
    """
    Manage edge dergradation methods
    """

    def calculate_adjustment_factor(
        self, usage_impact: float, age_impact: float
    ) -> float:
        """
        Calculate adjustment factor
        """
        return 1 + (self.coeff_usage * usage_impact) + (self.coeff_age * age_impact)

    def adjustment_factor(self, ids: list) -> float:
        """
        Return edge *adjustment_factor*
        """
        return self.calculate_adjustment_factor(
            usage_impact=self.get_usage_impact(ids=ids),
            age_impact=self.get_age_impact(ids=ids),
        )

    def calculate_adjusted_length(
        self, length: float, adjustment_factor: float
    ) -> float:
        """
        Calculate edge *adjusted_length*
        """
        return length * adjustment_factor

    def adjusted_length(self, ids: list) -> float:
        """
        Return edge *adjusted_length*
        """
        return self.calculate_adjusted_length(
            length=self.get_length(ids=ids),
            adjustment_factor=self.adjustment_factor(ids=ids),
        )

    def update_adjusted_length(self, ids: list):
        """
        Update *adjusted_length* value
        """
        adjusted_length = self.adjusted_length(ids=ids)
        self.set_adjusted_length(ids=ids, value=adjusted_length)

    def top_degraded_edges(self, percent: float = 0):
        """
        Filter most degradaded edges by their length percentage
        """
        if percent > 100:
            raise ValueError("enter a value between 0 and 100")
        edges_ids = self.streets
        total_length = 0
        edges_info = []
        for edge_ids in edges_ids:
            length = self.get_edge_attribute(ids=edge_ids, attribute="length")
            edge_info = {
                "ids": edge_ids,
                "degradation": self.adjustment_factor(ids=edge_ids),
                "length": length,
            }
            edges_info.append(edge_info)
            total_length += length
        sorted_edges_info = sorted(
            edges_info, key=lambda x: x["degradation"], reverse=True
        )
        remaining_length = (percent / 100) * total_length
        result = []
        for edge_info in sorted_edges_info:
            remaining_length -= edge_info["length"]
            if remaining_length < 0:
                break
            else:
                result.append(edge_info["ids"])
        return result



"""
.. module:: piperabm.infrastructure.query.add
:synopsis: Add new network elements.
"""

import numpy as np
from copy import deepcopy

from piperabm.tools.coordinate import distance as ds


class Add:
    """
    Add new network elements
    """

    def check_id(self, id):
        """
        Check whether id already exists
        """
        if id is None:
            id = self.new_id()
        else:
            if id in self.nodes:
                id = self.new_id()
                print("id already exists. replaced with new id.")
        return id

    def new_id(self):
        """
        Generate new unique random id
        """
        return int(
            np.random.randint(low=0, high=np.iinfo(np.int64).max, dtype=np.int64)
        )

    def add_junction(
        self, pos: list, id: int = None, name: str = "", report: bool = False
    ):
        """
        Add junction node.
        These are the nodes that connect edges in the network and represent a physical point in the world.

        Parameters
        ----------
        pos : list
            A list of [x, y] coordinates showing the position in space.
        id : int, optional
            The unique id number. The default is `None`, and if stays `None`, system will automatically assign a new unique id.
        name : str, optional
            Optional name of the element.
        report : bool
            If `True`, system will report successful creation of this element.
        """
        type = "junction"
        id = self.check_id(id)
        self.G.add_node(id, name=name, type=type, x=float(pos[0]), y=float(pos[1]))
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id

    def add_home(self, pos: list, id: int = None, name: str = "", report: bool = False):
        """
        Add home node.
        These are the nodes where agents live and belong to. Agents from the same home are a family and together they form a household.

        Parameters
        ----------
        pos : list
            A list of [x, y] coordinates showing the position in space.
        id : int, optional
            The unique id number. The default is `None`, and if stays `None`, system will automatically assign a new unique id.
        name : str, optional
            Optional name of the element.
        report : bool
            If `True`, system will report successful creation of this element.
        """
        type = "home"
        id = self.check_id(id)
        self.G.add_node(id, name=name, type=type, x=float(pos[0]), y=float(pos[1]))
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id

    def add_market(
        self,
        pos: list,
        resources: dict = {
            "food": 10,
            "water": 10,
            "energy": 10,
        },
        enough_resources: dict = None,
        id: int = None,
        name: str = "",
        report: bool = False,
    ):
        """
        Add market node
        These are the nodes where resources are bought and sold. The influx of resources to the model only happens through markets. They also act as social hubs in the model.

        Parameters
        ----------
        pos : list
            A list of [x, y] coordinates showing the position in space.
        resources: dict
            A dictionary showing how much of resources does the market have in stock.
        enough_resources: dict
            This shows the maximum amount of each resource market can have in stock. The default value is `None`. If stays `None`, system will rewrite it with a dict equal to `resources` dict.
        id : int, optional
            The unique id number. The default is `None`, and if stays `None`, system will automatically assign a new unique id.
        name : str, optional
            Optional name of the element.
        report : bool
            If `True`, system will report successful creation of this element.
        """
        type = "market"
        id = self.check_id(id)
        resource_kwargs = {}
        if enough_resources is None:
            enough_resources = {}
            for resource_name in self.resource_names:
                enough_resources[resource_name] = None
        for resource_name in self.resource_names:
            if enough_resources[resource_name] is None:
                enough_resources[resource_name] = deepcopy(resources[resource_name])
            resource_kwargs[resource_name] = resources[resource_name]
            resource_kwargs["enough_" + resource_name] = enough_resources[resource_name]

        self.G.add_node(
            id,
            name=name,
            type=type,
            x=float(pos[0]),
            y=float(pos[1]),
            balance=0,
            **resource_kwargs,
        )

        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} node at position {pos} added.")
        return id

    def add_street(
        self,
        pos_1: list,
        pos_2: list,
        name: str = "",
        usage_impact: float = 0,
        age_impact: float = 0,
        report: bool = False,
    ):
        """
        Add street edge. These edges are used by agents to move around the simulation world.

        Parameters
        ----------
        pos_1 : list
            A list of [x, y] coordinates showing the position of one of the ends in space.
        pos_2 : list
            A list of [x, y] coordinates showing the position of the other end in space.
        name : str, optional
            Optional name of the element.
        usage_impact : float, optional
            The more the element is used, this number will grow. It will be used to calcualte `adjustment_facor` when computing the degradation. The default is 0.
        age_impact : float, optional
            The more the element age, this number will grow. It will be used to calcualte `adjustment_facor` when computing the degradation. The default is 0.
        report : bool
            If `True`, system will report successful creation of this element.
        """
        type = "street"
        id_1 = self.add_junction(pos=pos_1)
        id_2 = self.add_junction(pos=pos_2)
        length = ds.point_to_point(pos_1, pos_2)
        adjustment_factor = self.calculate_adjustment_factor(
            usage_impact=usage_impact, age_impact=age_impact
        )
        adjusted_length = self.calculate_adjusted_length(
            length=length, adjustment_factor=adjustment_factor
        )
        self.G.add_edge(
            id_1,
            id_2,
            name=name,
            length=length,
            adjusted_length=adjusted_length,
            usage_impact=usage_impact,
            age_impact=age_impact,
            type=type,
        )
        self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(f">>> {type} edge at positions {pos_1}-{pos_2} added.")
        return id

    def add_neighborhood_access(
        self,
        id_1: list,
        id_2: list,
        name: str = "",
        usage_impact: float = 0,
        age_impact: float = 0,
        report: bool = False,
    ):
        """
        Add neighborhood access edge. These edges connect homes and markets to the street network, allowing agents to access these nodes.

        Parameters
        ----------
        pos_1 : list
            A list of [x, y] coordinates showing the position of one of the ends in space.
        pos_2 : list
            A list of [x, y] coordinates showing the position of the other end in space.
        name : str, optional
            Optional name of the element.
        usage_impact : float, optional
            The more the element is used, this number will grow. It will be used to calcualte `adjustment_facor` when computing the degradation. The default is 0.
        age_impact : float, optional
            The more the element age, this number will grow. It will be used to calcualte `adjustment_facor` when computing the degradation. The default is 0.
        report : bool
            If `True`, system will report successful creation of this element.
        """
        type = "neighborhood_access"
        length = ds.point_to_point(self.get_pos(id_1), self.get_pos(id_2))
        adjustment_factor = self.calculate_adjustment_factor(
            usage_impact=usage_impact, age_impact=age_impact
        )
        adjusted_length = self.calculate_adjusted_length(
            length=length, adjustment_factor=adjustment_factor
        )
        self.G.add_edge(
            id_1,
            id_2,
            name=name,
            length=length,
            adjusted_length=adjusted_length,
            usage_impact=usage_impact,
            age_impact=age_impact,
            type=type,
        )
        # self.baked_streets = False
        self.baked_neighborhood = False
        if report is True:
            print(
                f">>> {type} edge at positions {self.get_pos(id_1)} - {self.get_pos(id_2)} added."
            )
        return id



"""
Solves a single resource for multiple players at fixed price
"""

from scipy.optimize import minimize
import warnings

from piperabm.economy.accessibility import accessibility


warnings.filterwarnings(
    "ignore",
    message="Values in x were outside bounds during a minimize step, clipping to bounds",
)


class NashBargaining:

    def transactions(players: list, price: float) -> dict:
        """
        Calculate the list of transfers for proper resource allocation
        """
        num_players = len(players)

        if num_players == 0:
            return {"resource": [], "money": []}

        def objective(resource_transafers):
            """
            Optimization objective function
            """
            accessibilities = [
                accessibility(
                    players[i]["resource"] + resource_transafers[i],
                    players[i]["enough_resource"],
                )
                for i in range(num_players)
            ]
            result = 1
            for a in accessibilities:
                result *= a
            return -result

        initial_guess = []
        bounds = []
        for i in range(num_players):
            need = max(0, players[i]["enough_resource"] - players[i]["resource"])
            possible = players[i]["balance"] / price
            upper_bound = min(need, possible)
            lower_bound = -players[i]["resource"]
            bound = (lower_bound, upper_bound)
            bounds.append(bound)
            initial_guess.append((upper_bound + lower_bound) / 2)
        constraints = [
            {"type": "eq", "fun": lambda resource_transafers: sum(resource_transafers)},
        ]
        result = minimize(
            objective,
            initial_guess,
            bounds=bounds,
            constraints=constraints,
        )
        resource_transfers = [
            float(resource_transfer) for resource_transfer in result.x
        ]
        money_transfers = [
            float(-resource_transfer * price)
            for resource_transfer in resource_transfers
        ]
        return {"resource": resource_transfers, "money": money_transfers}

    def apply(players: list, transactions: dict) -> list:
        """
        Apply transfers to players
        """
        num_players = len(players)
        for i in range(num_players):
            players[i]["balance"] = players[i]["balance"] + transactions["money"][i]
            players[i]["resource"] = (
                players[i]["resource"] + transactions["resource"][i]
            )
        return players


if __name__ == "__main__":
    price = 10
    player_1 = {
        "id": 1,
        "type": "agent",
        "resource": 9,
        "enough_resource": 10,
        "balance": 100,
    }
    player_2 = {
        "id": 2,
        "type": "agent",
        "resource": 6,
        "enough_resource": 10,
        "balance": 100,
    }
    player_3 = {
        "id": 3,
        "type": "agent",
        "resource": 3,
        "enough_resource": 10,
        "balance": 10,
    }
    players = [player_1, player_2, player_3]

    # Initial
    print("\n" + ">>> " + "Initial: ")
    for player in players:
        print(player)

    # Solve
    transactions = NashBargaining.transactions(players, price)
    players = NashBargaining.apply(players, transactions)

    # Final
    print("\n" + ">>> " + "Final: ")
    for player in players:
        print(player)



from piperabm.economy.trade.trade_solver import MultiResourceTrade as mrt


prices = {"food": 10, "water": 10, "energy": 10}
player_1 = {
    "id": 1,
    "type": "market",
    "resources": {"food": 100, "water": 100, "energy": 100},
    "enough_resources": {"food": 100, "water": 100, "energy": 100},
    "balance": 0,
}
player_2 = {
    "id": 2,
    "type": "agent",
    "resources": {"food": 10, "water": 3, "energy": 10},
    "enough_resources": {"food": 10, "water": 10, "energy": 10},
    "balance": 100,
}
player_3 = {
    "id": 3,
    "type": "agent",
    "resources": {"food": 1, "water": 2, "energy": 3},
    "enough_resources": {"food": 10, "water": 10, "energy": 10},
    "balance": 10,
}
players = [player_1, player_2, player_3]

players = mrt.solve(players=players, prices=prices)


if __name__ == "__main__":
    for player in players:
        print(f"resources: {player['resources']}, balance: {player['balance']}")



import os
from copy import deepcopy
import matplotlib.pyplot as plt

from piperabm.tools.average import average as avg


class TravelDistance:
    """
    Manage travel distance measurement
    """

    type = "travel distance"

    def __init__(self, measurement):
        self.measurement = measurement
        self.values = []

    def add(self, value: float) -> None:
        """
        Add new travel distance value.
        """
        self.values.append(value)

    def filter(self, _from=None, _to=None):
        """
        Filter values in a specific range.

        Parameters
        ----------
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(self.values)
        return self.values[_from:_to]

    def __call__(self, _from=None, _to=None):
        return self.filter(_from=_from, _to=_to)

    def average(self, _from=None, _to=None) -> float:
        """
        Calculate total average.

        Parameters
        ----------
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        values = self.__call__(_from=_from, _to=_to)
        weights = self.measurement.delta_times(_from=_from, _to=_to)
        result = avg.arithmetic(values=values, weights=weights)
        if isinstance(result, complex):
            result = float(result.real)
        return result

    def create_plot(self, _from=None, _to=None, info=None):
        """
        Create plot.

        Parameters
        ----------
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        """
        fig, ax = plt.subplots()
        title = "Travel Distance"
        ylabel = deepcopy(title)
        if info is not None:
            title += info
        ax.set_title(title)
        xs = self.measurement.filter_times(_from=_from, _to=_to)
        yx = self.filter(_from=_from, _to=_to)
        ax.plot(xs, yx, color="blue")
        ax.set_xlabel("Time")
        ax.set_ylabel(ylabel)
        return fig

    def show(self, _from=None, _to=None, info: str = None):
        """
        Draw plot.

        Parameters
        ----------
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        info: str, default=None
            The extra information shown in the plot.
        """
        fig = self.create_plot(_from=_from, _to=_to, info=info)
        plt.show()

    def save(self, _from=None, _to=None, info=None):
        """
        Save plot.

        Parameters
        ----------
        _from: int, default=None
            The step number in time to start summation. If `None`, the earliest step will be considered.
        _to: int, default=None
            The step number in time to end summation. If `None`, the latest step will be considered.
        info: str, default=None
            The extra information shown in the plot.
        """
        fig = self.create_plot(_from=_from, _to=_to, info=info)
        path = self.measurement.result_directory
        filepath = os.path.join(path, self.type)
        fig.savefig(filepath)

    def serialize(self) -> dict:
        """
        Serialize.
        """
        return {"values": self.values, "type": self.type}

    def deserialize(self, data: dict) -> None:
        """
        Deserialize.
        """
        self.values = data["values"]


if __name__ == "__main__":

    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour)  # Base

    # 1
    measure.add_time(value=1 * hour)
    measure.travel_distance.add(value=1.1)
    # 2
    measure.add_time(value=2 * hour)
    measure.travel_distance.add(value=0.9)
    # 3
    measure.add_time(value=3 * hour)
    measure.travel_distance.add(value=0.3)
    # 4
    measure.add_time(value=4 * hour)
    measure.travel_distance.add(value=0.46)
    # 5
    measure.add_time(value=5 * hour)
    measure.travel_distance.add(value=0.2)

    _from = None
    _to = None
    print("travel distances: ", measure.travel_distance(_from=_from, _to=_to))
    measure.travel_distance.show(_from=_from, _to=_to)



"""
.. module:: piperabm.infrastructure.serialize
:synopsis: Infrastructure serialization mixin providing serialize/deserialize methods for the Infrastructure class.
"""

from piperabm.tools.nx_serializer import nx_serialize, nx_deserialize


class Serialize:
    """
    Serialization methods
    """

    def serialize(self):
        """
        Serialize
        """
        data = {}
        data["G"] = nx_serialize(self.G)
        data["coeff_usage"] = self.coeff_usage
        data["coeff_age"] = self.coeff_age
        data["baked_streets"] = self.baked_streets
        data["baked_neighborhood"] = self.baked_neighborhood
        data["heuristic_paths"] = self.heuristic_paths.serialize()
        data["type"] = self.type
        return data

    def deserialize(self, data):
        """
        Deserialize
        """
        self.G = nx_deserialize(data["G"])
        self.coeff_usage = data["coeff_usage"]
        self.coeff_age = data["coeff_age"]
        self.baked_streets = data["baked_streets"]
        self.baked_neighborhood = data["baked_neighborhood"]
        self.heuristic_paths.deserialize(data["heuristic_paths"])


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_0 import model

    print(model.infrastructure.serialize())



"""
Manage simulation execution: stepping the model forward and handling persistence.

This mix-in provides:
 - run(): loop over multiple time steps (or until all agents die),
   with optional save/resume behavior.
 - update(): perform a single time‐step update, including trading,
   agent/infrastructure updates, and optional delta‐based serialization.
"""

from copy import deepcopy
import keepdelta as kd

from piperabm.model.trade import Trade
from piperabm.tools.json_file import JsonFile


class Update(Trade):
    """
    Manage running simulation

    Inherits from Trade, so it has access to self.trade(…) for market transactions.
    """

    def run(
        self,
        n: int = None,
        step_size: float = 3600,
        save: bool = False,
        save_transactions: bool = False,
        resume: bool = False,
        report: bool = False,
    ):
        """
        Run model for multiple steps.

        Parameters
        ----------
        n : int or None
            Number of steps to run the model. If set to `None`, model keeps running until no agents remain alive in the society.
        step_size : float
            Time increment for each step (e.g. seconds). If set to large values, the model runs faster but the model may not be able to capture some of the interactions.
        save : bool
            If `True`, saves the results buy serializing state snapshots and creating deltas and recording them to disk. The initial state of model (named `initial.json`), the final state of the model (named `final.json`) as well as the changes (deltas) during each step (named `simulation.json`) can be found in the `result` folder in the working directory.
        save_transactions : bool
            If `True`, record all transactions to disk. The file will be named `transaction.csv` and will be located in the `result` folder in the working directory.
        resume : bool
            If `True`, attempt to resume from last saved state instead of starting fresh.
        report : bool
            If `True`, print progress for n-step runs.
        """
        # Remove previous save file if exists
        if self.path is not None:
            path = self.result_directory
            if resume is False:
                # Remove deltas
                simulation_file = JsonFile(path, "simulation")
                simulation_file.remove()
                # Remove final state
                final_file = JsonFile(path, "final")
                final_file.remove()
                # Load initial state if exists
                initial_file = JsonFile(path, "initial")
                if initial_file.exists():
                    self.load_initial()
                else:
                    if save is True:
                        self.save_initial()
            else:
                # Load final state if exists
                final_file = JsonFile(path, "final")
                if final_file.exists() is True:
                    self.load_final()
                else:
                    # Load initial state if final state doesn't exists
                    initial_file = JsonFile(path, "initial")
                    if initial_file.exists() is True:
                        self.load_initial()
                        # Apply deltas if exists
                        simulation_file = JsonFile(path, "simulation")
                        if simulation_file.exists() is True:
                            self.apply_deltas()

        # Run until all agents die
        if n is None:
            while True:
                self.update(
                    duration=step_size, save=save, save_transactions=save_transactions
                )
                if len(self.society.alive_agents) == 0:
                    break

        # Run for certain steps
        else:
            for i in range(n):
                if report is True:
                    print(f"Progress: {(i + 1) / n * 100:.1f}% complete")
                self.update(
                    duration=step_size, save=save, save_transactions=save_transactions
                )

    def update(
        self, duration: float, save: bool = False, save_transactions: bool = False
    ):
        """
        Update model for a single step.

        Steps performed:
        1. (If save) snapshot current state for delta comparison.
        2. Execute trades in each market and household.
        3. Update agent behaviors and infrastructure effects.
        4. Reset market balances and refill resources.
        5. Increment time and step counters.
        6. (If save) compute & append state delta, save final state.
        7. (If save_transactions) append transaction list to log.
        """
        # Delta
        if save is True:
            # Create current state
            previous_serialized = deepcopy(self.serialize())

        # Trade
        transactions = []
        for market_id in self.infrastructure.markets:  # Agents in market
            agents = self.society.agents_in(id=market_id)
            if len(agents) >= 1:
                transactions += self.trade(agents=agents, markets=[market_id])
        for home_id in self.infrastructure.homes:  # Agents in home
            agents = self.society.agents_in(id=home_id)
            if len(agents) >= 2:
                transactions += self.trade(agents=agents)
        # transactions
        for transaction in transactions:
            transaction.append(self.time)

        # Agents activity impact
        self.society.update(duration)

        # Climate impact
        self.infrastructure.update(duration)

        # Charge Markets (resource influx)
        markets = self.infrastructure.markets
        for id in markets:
            # Reset balance
            self.infrastructure.set_balance(id=id, value=0)
            # Reset resources
            for name in self.resource_names:
                enough_amount = self.infrastructure.get_enough_resource(
                    id=id, name=name
                )
                self.infrastructure.set_resource(id=id, name=name, value=enough_amount)

        # General
        self.step += 1
        self.time += duration

        # Delta
        if save is True:
            # Create new current state and compare it to the previous one
            current_serialized = self.serialize()
            delta = kd.create(old=previous_serialized, new=current_serialized)
            self.append_delta(delta)
            self.save_final()

        # Transactions
        if save_transactions is True:
            self.append_transactions(transactions)

        return transactions



"""
This example generates 1000 points inside a two adjacent triangles
"""

import matplotlib.pyplot as plt

from piperabm.tools.mesh import Triangle, Patch


pos_1 = [0, 0]
pos_2 = [0, 2]
pos_3 = [2, 0]
pos_4 = [1, 1]
pos_5 = [2, 1]
triangle_1 = Triangle(pos_1, pos_2, pos_3)
triangle_2 = Triangle(pos_3, pos_4, pos_5)

patch = Patch()
patch.add(triangle_1, triangle_2)

xs = []
ys = []
for _ in range(1000):
    point = patch.random_point()
    xs.append(point[0])
    ys.append(point[1])

plt.scatter(xs, ys)  # Plot points
plt.show()



import random

from piperabm.tools.mesh.triangle import Triangle


class Patch:
    """
    A group of triangles create a patch.
    """

    def __init__(self):
        self.library = []

    def add(self, *triangles):
        """
        Add new triangle objects to the patch.
        """
        for triangle in triangles:
            if isinstance(triangle, Triangle):
                self.library.append(triangle)

    @property
    def weights(self) -> list:
        """
        Return a list of weights oeganized by triangle orders.
        """
        result = []
        for triangle in self.library:
            result.append(triangle.weight)
        return result

    @property
    def indexes(self) -> list:
        """
        Return the list of triangle indexes.
        """
        result = []
        for i in range(len(self.library)):
            result.append(i)
        return result

    def random_point(self):
        """
        Generate a random point inside the patch based on the weights.
        """
        values = self.indexes
        weights = self.weights
        index = random.choices(values, weights=weights, k=1)[0]
        triangle = self.library[index]
        return triangle.random_point()


if __name__ == "__main__":
    pos_1 = [0, 0]
    pos_2 = [0, 2]
    pos_3 = [2, 0]
    pos_4 = [1, 1]
    pos_5 = [2, 1]
    triangle_1 = Triangle(pos_1, pos_2, pos_3)
    triangle_2 = Triangle(pos_3, pos_4, pos_5)

    patch = Patch()
    patch.add(triangle_1, triangle_2)

    print(patch.weights)



"""
This example generates 1000 points inside a single triangle
"""

import matplotlib.pyplot as plt

from piperabm.tools.mesh import Triangle


pos_1 = [0, 0]
pos_2 = [2, 0]
pos_3 = [0, 1]
triangle = Triangle(pos_1, pos_2, pos_3)

xs = []
ys = []
for _ in range(1000):
    point = triangle.random_point()
    xs.append(point[0])
    ys.append(point[1])

plt.scatter(xs, ys)  # Plot points
plt.show()



"""
.. module:: piperabm.infrastructure.query.query
:synopsis: Query network elements.
"""

import networkx as nx
import numpy as np

from piperabm.tools.coordinate import distance as ds
from piperabm.infrastructure.query.add import Add
from piperabm.infrastructure.query.get import Get
from piperabm.infrastructure.query.set import Set


class Query(Add, Get, Set):
    """
    Query network elements
    """

    def has_node(self, id: int) -> bool:
        """
        Check whether the network already contains the node
        """
        return self.G.has_node(id)

    def has_edge(self, ids: list) -> bool:
        """
        Check whether the network already contains the edge
        """
        return self.G.has_edge(*ids)

    def is_isolate(self, id: int) -> bool:
        """
        Check if the node is isolated
        """
        return nx.is_isolate(self.G, id)

    '''
    def filter_nodes_closer_than(self, id: int, distance: float, nodes: list = None) -> list:
        """
        Filter *nodes* that are within the *distance* from *id*
        """
        result = []
        if nodes is None:
            nodes = self.nodes
        for node_id in nodes:
            if distance >= self.heuristic_paths.estimated_distance(id_start=id, id_end=node_id):
                result.append(node_id)
        return result
    '''

    def replace_node(self, id: int, new_id: int, report: int = False) -> None:
        """
        Replace a node with another node
        """
        # Find all adjacent edges
        edges_ids = self.edges_from(id)
        # Apply change to adjacent edges
        for edge_ids in edges_ids:
            # Create new edge
            if edge_ids[0] == id:
                new_edge_ids = [new_id, edge_ids[1]]
            else:
                new_edge_ids = [edge_ids[0], new_id]
            data = self.get_edge_attributes(ids=edge_ids)
            data["length"] = ds.point_to_point(
                self.get_pos(new_edge_ids[0]), self.get_pos(new_edge_ids[1])
            )
            adjustment_factor = self.calculate_adjustment_factor(
                usage_impact=data["usage_impact"], age_impact=data["age_impact"]
            )
            data["adjusted_length"] = self.calculate_adjusted_length(
                length=data["length"],
                adjustment_factor=adjustment_factor,
            )
            self.G.add_edge(new_edge_ids[0], new_edge_ids[1], **data)
            if report is True:
                print(
                    f">>> {type} edge at positions {self.get_pos(new_edge_ids[0])} - {self.get_pos(new_edge_ids[1])} added."
                )
            # Remove old edge
            self.remove_edge(ids=edge_ids, report=report)
        # Remove old node
        self.remove_node(id, report=report)

    def impact(self, edges: list = []):
        """
        Impact the network by removing a list of edges
        """
        if self.baked is False:
            print("First bake the model")
            raise ValueError
        for ids in edges:
            self.remove_edge(ids=ids)
            id_1 = ids[0]
            id_2 = ids[1]
            if self.get_node_type(id=id_1) == "junction" and self.is_isolate(id=id_1):
                self.remove_node(id=id_1)
            if self.get_node_type(id=id_2) == "junction" and self.is_isolate(id=id_2):
                self.remove_node(id=id_2)

    def random_edges(self, percent: float = 0):
        """
        Filter random edges by their length percentage
        """
        if percent > 100:
            raise ValueError("enter a value between 0 and 100")
        edges_ids = self.streets
        total_length = 0
        edges_info = []
        for edge_ids in edges_ids:
            length = self.get_edge_attribute(ids=edge_ids, attribute="length")
            total_length += length
            edge_info = {"ids": edge_ids, "length": length}
            edges_info.append(edge_info)
        remaining_length = (percent / 100) * total_length
        result = []
        np.random.shuffle(edges_info)
        for edge_info in edges_info:
            remaining_length -= edge_info["length"]
            if remaining_length < 0:
                break
            else:
                result.append(edge_info["ids"])
        return result

    def remove_edge(self, ids: list = None, report: bool = False):
        """
        Remove edge
        """
        if report is True:
            print(
                f">>> {self.get_edge_type(ids=ids)} edge at {self.get_pos(ids[0])} - {self.get_pos(ids[1])} removed."
            )
        self.G.remove_edge(*ids)

    def remove_node(self, id: int, report: bool = False):
        """
        Remove node
        """
        if report is True:
            print(
                f">>> {self.get_node_type(id=id)} node at position {self.get_pos(id)} removed."
            )
        self.G.remove_node(id)

    def nodes_closer_than(
        self,
        id: int,
        search_radius: float = 0,
        nodes: list = None,
        include_self: bool = False,
    ):
        """
        Filter *nodes* that are within the *distance* from *id*
        """
        if nodes is None:
            nodes = self.nodes
        result = []
        for node_id in nodes:
            if search_radius >= self.heuristic_paths.estimated_distance(
                id_start=id, id_end=node_id
            ):
                if include_self is False:
                    if node_id != id:
                        result.append(node_id)
                else:
                    result.append(node_id)
        return result

    @property
    def junctions(self) -> list:
        """
        Return all junction nodes
        """
        try:
            return [
                n
                for n, attr in self.G.nodes(data=True)
                if attr.get("type") == "junction"
            ]
        except:
            return []

    @property
    def nonjunctions(self) -> list:
        """
        Return all nonjunction nodes
        """
        try:
            return [
                n
                for n, attr in self.G.nodes(data=True)
                if attr.get("type") != "junction"
            ]
        except:
            return []

    @property
    def homes(self) -> list:
        """
        Return all homes nodes
        """
        try:
            return [
                n for n, attr in self.G.nodes(data=True) if attr.get("type") == "home"
            ]
        except:
            return []

    @property
    def markets(self) -> list:
        """
        Return all market nodes
        """
        try:
            return [
                n for n, attr in self.G.nodes(data=True) if attr.get("type") == "market"
            ]
        except:
            return []

    @property
    def streets(self) -> list:
        """
        Return all street edges
        """
        try:
            return [
                (u, v)
                for u, v, attr in self.G.edges(data=True)
                if attr.get("type") == "street"
            ]
        except:
            return []

    @property
    def neighborhood_accesses(self) -> list:
        """
        Return all neighborhood access edges
        """
        try:
            return [
                (u, v)
                for u, v, attr in self.G.edges(data=True)
                if attr.get("type") == "neighborhood_access"
            ]
        except:
            return []
